:- module( mlu, [
	k_fold_learn/4,
	k_fold_pairwise_comparisons/6,
	k_fold_pairwise_predictions/6,
	mlu_errors/0,
	mlu_version/2
		] ).

:- use_module( library(requires) ).

:- lib(os_lib).       % canonical of lib(os)
:- lib(debug_call).
:- lib(pack_errors).



/** <module> Machine learning utilities

A menagerie of machine learning utilities.

Currently implements k_fold learning and k_fold comparative performance plots via Real.

It is likely that bootstrapping will be added soon and also a couple of additional types of comparative plots.

@author    nicos angelopoulos
@version   0.1,   2016/3/5
@see http://stoics.org.uk/~nicos/sware/mlu
@see http://stoics.org.uk/~nicos/sware/mlu/html/mlu.html
@see pack(mlu/examples/stoic.pl)

*/

/** mlu_version( -Version, -Date ).

Current version and release date for the library.

*/
mlu_version( 0:1:0, date(2016,3,5) ).


:- use_module( library(random) ).  % random_permutation/2.

:- suggests(mtx).  				% does not complain if not available.


k_fold_learn_defaults( Defs ) :-
	Defs  = [  has_header(true), is_matrix(IsM),
			 folds(10),        repeats(1),
			 predictor(false), include_header(false), 
			 include_stem(arity),
			 call_options(false),
			 statistics(_),    segments(_)
	        ],
	holds( current_predicate(mtx/2), IsM ).

/** k_fold_learn( +Data, +Goal, ?Models, +Opts ).

	Split data to N segments and run Goal on each one to learn
	a Model or a sequence of models. Data should be a list of things.
	Normally it will be a list of data rows, such as one read by 
	csv_read_file/3. The predicate can use existing partitions of Data 
	(see option segments(Segms)).

	Opts 
	 * call_options(Copts=false)
	   if anything else than false, then Copts is added as last argument to Goal
	   (after it is passed through en_list/2). The option fold(K) is also added
	   where K is the number of fold.

	 * folds(Folds=10)
	   number of segments to split the data to 
	   (currently only exhaustive and mutual exclusive splits are supported)

	 * has_header(HasHdr=true)
	   whether Data include a Header element (first)

	 * include_header(IncHdr=false)
	   include header to segments, if HasHdr==true and IncHdr==false 'false' is added as a header

	 * include_stem(Stem=arity)
	   include automatically generated stem to call. The stem looks like,
	   <learner>_<LSeq>_<FSeq>[_<RSeq>. eg truism_A_01_r1
	   * arity
	     include if there is definition for Goal+1 (accounts for Copts too)

	   * false
	     does not add Stem no matter what

	   * model
	     Stem replaces to model argument, 

	   * true
	     adds it no matter what

	 * is_matrix(IsM=true/false)
	   whether to pass the data through mtx/2. Default depends on whether mtx/2 
	   is in memory (true) or not (false)

	 * model_name_stem(Stem)
	   Stem for model name

      * predictor(Predictor=false)
        if non false, Predictor is called after calling a learner.
	   the Predictor takes Model and (hold out) Data as input and returns Statistic

	 * repeats(R=1)
	   how many times to run Goal on each fold. If 1, then Models is a list
	   of models as returned by goal, else, it is a list of lists of models-
	   each sublist corresponding the one repeat run on a segment

	 * segments(Segms)
	   returns the data segments as passed to the learner. if Segs is ground
	   these segments are used rather than a new partition

	 * statistics(Statistics)
	   one statistic for each application for the Predictor (nested as per Models)

==
?- k_fold_learn( [a,b,c], true, Mods, [is_matrix(false)] ).
ERROR: pack(mlu): Insufficient length of data (2) as 10 folds are requierd

?- assert(true(_,true)), assert(second(_,false)).

?- debug( mlu(k_fold_learn)), numlist( 1, 33, Tthree ),
   k_fold_learn( Tthree, true, Mods, [is_matrix(false),segments(Segms),repeats(10)]),
   length( Mods, LenMods ), 
   flatten( Mods, AllMods ), 
   length( AllMods, LenAllMods ).

% Data fold partition, segment lengths: [3,3,3,3,4,3,3,3,3,4]
LenMods = 10,
LenAllMods = 100.

==
@author nicos angelopoulos
@version  0.1 2015/12/17
@see  example in pack(mlu/examples/stoic.pl)

*/
k_fold_learn( MtxF, GoalPrv, Models, Args ) :-
	( GoalPrv = multi(GPsfxs,Goal) -> length(GPsfxs,GenM); Goal = GoalPrv, GenM is 0 ),
	options_append( k_fold_learn, Args, Opts ),
	memberchk( segments(Segms), Opts ),
	holds( var(Segms), VarSegms ),
	options( folds(K), Opts ),
	k_fold_cross_learn_segments( VarSegms, MtxF, K, Hdr, Segms, Opts ),
	options( call_options(CoptsOpt), Opts ),
	compare( meta, CoptsOptIsFalse, CoptsOpt, ==(false) ),
	k_fold_call_options( CoptsOptIsFalse, CoptsOpt, HasCopts, Copts ),
	options( repeats(R), Opts ),
	compare( meta, SingR, R, =:=(1) ),
	holds( 1 > K, ApplyTerm ),
	options( include_header(IncHdr), Opts ),
	options( predictor(PredPrv), Opts ),
	holds( PredPrv\==false, HasPred ),
	k_fold_moduled( PredPrv, Pred ),
	options( include_stem(IncStemPrv), Opts ),
	k_fold_include_stem( IncStemPrv, HasCopts, Goal, IncStem ),
	k_fold_moduled( Goal, Moal ),
	number_codes( K, Codes ),
	length( Codes, LenK ),

	( memberchk(stem(Stem),Opts) -> true;
			( var(GPsfxs) ->
				  goal_predicate_name(Moal,Stem) 
				  ;
				  goal_predicate_name(Moal,GStem),
				  findall( AStem, (member(Spfx,GPsfxs),atomic_list_concat([GStem,Spfx],'_',AStem)), Stem )
			)
	),
	k_fold_apply( ApplyTerm, 1, K, LenK, Segms, IncHdr, Hdr, SingR, R, HasPred, Pred, Moal, IncStem/HasCopts, Stem, Copts, Models, GenM, Stats, _Folds ),
	memberchk( statistics(Stats), Opts ).

k_fold_include_stem( arity, HasCopts, Goal, Exists ) :-
	k_fold_include_stem_has_call_options_arity_increase( HasCopts, HasCoptsNum ),
	( Goal = Mod:Noal -> true; Noal = Goal, Mod = user ),
	Noal =.. [Pname|Args],
	length( Args, Nrgs ),
	Sum is Nrgs + HasCoptsNum + 2 + 1,
	holds( current_predicate( Mod:Pname/Sum ), Exists ).
k_fold_include_stem( true, _HasCopts, _Goal, true ).
k_fold_include_stem( false, _HasCopts, _Goal, false ).
k_fold_include_stem( model, _HasCopts, _Goal, model ).

k_fold_include_stem_has_call_options_arity_increase(  true, 1 ).
k_fold_include_stem_has_call_options_arity_increase( false, 0 ).

k_fold_cross_learn_segments( true, MtxF, K, Hdr, Segms, Opts ) :-
	k_fold_mtx_header_rows( MtxF, Hdr, Rows, Opts ),
	random_permutation( Rows, Shuffled ),
	length( Rows, Dlen ),
	debug( mlu(k_fold_learn), 'Data rows length: ~d (excluding header if present) ', [Dlen] ),
	compare( meta, Op, Dlen, =<(K) ),
	k_fold_enough_data( Op, Dlen, K ),
	numlist( 0, Dlen, Idxs ),
	n_breaks( Idxs, K, [_|Breaks], [] ),
	debug( mlu(k_fold_learn), 'Data segment breaks: ~w', [Breaks] ),
	Breaks = [B|TBs],
	compare( meta, Cop, 1, >=(B) ),
	k_fold_break( Cop, 1, B, TBs, Shuffled, Segm, TSegms ),
	holds( debugging(mlu(k_fold_learn)), Dbg ),
	Segms = [Segm|TSegms],
	debug_segment_lengths( Dbg, Segms ).
k_fold_cross_learn_segments( false, MtxF, _K, Hdr, _Segms, Opts ) :-
	k_fold_mtx_header_rows( MtxF, Hdr, _Rows, Opts ).

k_fold_break( <>, I, _OldB, Bs, Data, [], Segms ) :-
	% k_fold_break_header( IncHdr, Hdr, Segm, Hegm ),
	k_fold_break_continue( Bs, I, Data, Segms ).
k_fold_break( =, I, B, TBs, [D|Data], [D|Segm], Segms ) :-
	J is I + 1,
	compare( meta, Cop, J, >=(B) ),
	k_fold_break( Cop, J, B, TBs, Data, Segm, Segms ).

k_fold_break_continue( [], _I, Data, [] ) :-
	k_fold_break_residual_data( Data ).
k_fold_break_continue( [B|TBs], I, Data, [Segm|Segms] ) :-
	% J is I + 1,
	compare( meta, Cop, I, >=(B) ),
	% should  always be Cop = (=) 
	debug( mlu(k_fold_learn), 'Breaking segment at position: ~d', I ),
	k_fold_break( Cop, I, B, TBs, Data, Segm, Segms ).

k_fold_apply( true, _I, _K, _LenK, _Segms, _IncH, _Hdr, _SingR, _R, _HasP, _Pctr, _Goal, _HasC, _StemIn, _Copts, [], _GenM, [], [] ).
k_fold_apply( false, I, K, LenK, Segms, IncH, Hdr, SingR, R, HasP, Pctr, Goal, HasCopts, StemIn, Copts, [SegMs|Ms], GenM, [SegStats|Stats], [Data-HoldOut|TFolds] ) :-
	nth1( I, Segms, Seg, Remainder ),
	flatter( Remainder, DataPrv ),
	k_fold_include_header( IncH, Hdr, DataPrv, Data ),
	k_fold_include_header( IncH, Hdr, Seg, HoldOut ),
	n_digits_min( LenK, I, Ipad ),
	( is_list(StemIn) ->
		findall( AStem, (member(AStemIn,StemIn),at_con([AStemIn,'_m',Ipad],'',AStem)), Stem )
		;
		at_con( [StemIn,'_m',Ipad], '', Stem )
	),
	k_fold_apply_segment( SingR, R, Goal, Data, HoldOut, HasCopts, Stem, Copts, HasP, Pctr, SegMs, GenM, SegStats ),
	J is I + 1,
	holds( J > K, Termin ),
	k_fold_apply( Termin, J, K, LenK, Segms, IncH, Hdr, SingR, R, HasP, Pctr, Goal, HasCopts, StemIn, Copts, Ms, GenM, Stats, TFolds ).

k_fold_include_header( true, Hdr, Rows, [Hdr|Rows] ).
k_fold_include_header( false, _Hdr, Rows, Rows ).

k_fold_apply_segment( =, 1, Goal, Seg, HoldOut, HasCopts, Stem, Copts, HasP, Pctr, SegMs, GenM, SegStats ) :-
	k_fold_call( HasCopts, Goal, Seg, HoldOut, Stem, Copts, HasP, Pctr, SegMs, GenM, SegStats ).
k_fold_apply_segment( <>, R, Goal, Seg, HoldOut, HasCopts, Stem, Copts, HasP, Pctr, SegMs, GenM, SegStats ) :-
	compare( meta, Op, R, =:=(0) ),
	k_fold_apply_segment_repeats( Op, R, Goal, Seg, HoldOut, HasCopts, Stem, Copts, HasP, Pctr, SegMs, GenM, SegStats ).

k_fold_apply_segment_repeats( =, 0, _Goal, _Seg, _Hout, _HasCopts, _Stem, _Copts, _HasP, _Pctr, _GenM, [], [] ).
k_fold_apply_segment_repeats( <>, R, Goal, Seg, Hout, HasCopts, StemIn, Copts, HasP, Pctr, [Mod|Mods], GenM, [St|Sts] ) :-
	at_con( [StemIn,'_',r,R], '', Stem ),
	k_fold_call( HasCopts, Goal, Copts, Seg, Hout, HasP, Pctr, Mod, GenM, St ),
	Q is R - 1,
	compare( meta, Op, Q, =:=(0) ),
	k_fold_apply_segment_repeats( Op, Q, Goal, Seg, Hout, HasCopts, Stem, Copts, HasP, Pctr, Mods, GenM, Sts ).

% k_fold_call( IncStem/IncOpts, Copts, Goal, Segm, Hout, HasP, Pctr, Model, Statistic ) :-
k_fold_call( true/true, Goal, Segm, Hout, Stem, Copts, HasP, Pctr, Model, GenM, Statistic ) :-
	!,
	call( Goal, Segm, Stem, Model, Copts ),
	k_fold_call_predictor( HasP, Pctr, Model, Hout, GenM, Statistic ).
k_fold_call( true/false, Goal, Segm, Hout, Stem, _Copts , HasP, Pctr, Model, GenM, Statistic ) :-
	!,
	call( Goal, Segm, Stem, Model ),
	k_fold_call_predictor( HasP, Pctr, Model, Hout, GenM, Statistic ).
k_fold_call( false/true, Goal, Segm, Hout, _Stem, Copts, HasP, Pctr, Model, GenM, Statistic ) :-
	!,
	call( Goal, Segm, Model, Copts ),
	k_fold_call_predictor( HasP, Pctr, Model, Hout, GenM, Statistic ).
k_fold_call( false/false, Goal, Segm, Hout, _Stem, _Copts, HasP, Pctr, Model, GenM, Statistic ) :-
	!,
	call( Goal, Segm, Model ),
	k_fold_call_predictor( HasP, Pctr, Model, Hout, GenM, Statistic ).
k_fold_call( model/true, Goal, Segm, Hout, Stem, Copts, HasP, Pctr, Stem, GenM, Statistic ) :-
	!,
	call( Goal, Segm, Stem, Copts ),
	k_fold_call_predictor( HasP, Pctr, Stem, Hout, GenM, Statistic ).
k_fold_call( model/false, Goal, Segm, Hout, Stem, _Copts, HasP, Pctr, Stem, GenM, Statistic ) :-
	!,
	call( Goal, Segm, Stem ),
	k_fold_call_predictor( HasP, Pctr, Stem, Hout, GenM,  Statistic ).

k_fold_call_predictor( true, Predictor, Model, Hout, GenM, Statistic ) :-
	k_fold_call_predictor_multi( GenM, Predictor, Model, Hout, Statistic ).
k_fold_call_predictor( false, _Predictor, _Model, _Hout, _GenM, false ). % fixme when GenM generate Genm false list

k_fold_call_predictor_multi( 0, Predictor, Model, Hout, Statistic ) :-
	!,
	call( Predictor, Model, Hout, Statistic ).
k_fold_call_predictor_multi( M, Predictor, Models, Hout, Statistics ) :-
	length( Models, LenMs ),
	( LenMs =:= M -> true;  throw( missmatch(number_of_models_generated(M,LenMs)) ) ),
	findall( Statistic, (member(Model,Models), call(Predictor,Model,Hout,Statistic) ), Statistics ).

k_fold_call_options( <>, CoptsOpt, true, Copts ) :-
	en_list( CoptsOpt, Copts ).
k_fold_call_options( =, false, false, false ).

k_fold_break_residual_data( [] ).
k_fold_break_residual_data( [D|Data] ) :-
	length( [D|Data], Dlen ),
	throw( pack_error(mlu,fold_data_residual(Dlen)) ).

k_fold_mtx_header_rows( MtxF, Hdr, Rows, Opts ) :-
	options( is_matrix(IsM), Opts ),
	k_fold_mtx( IsM, MtxF, Mtx ),
	options( has_header(HasHdr), Opts ),
	k_fold_header_data( HasHdr, Mtx, Hdr, Rows ).

k_fold_enough_data( =, _Dlen, _N ).
k_fold_enough_data( <>, Dlen, N ) :-
	throw( pack_error(mlu,fold_data_insufficient(Dlen,N)) ).

k_fold_header_data( true, [Hdr|Rows], Hdr, Rows ).
k_fold_header_data( false, Mtx, false, Mtx ).

k_fold_mtx( true, MtxF, Mtx ) :-
	mtx( MtxF, Mtx ).
k_fold_mtx( false, Mtx, Mtx ).

flatter( [], [] ).
flatter( [H|T], Data ) :-
	append( H, Var, Data ),
	flatter( T, Var ).

/*
flatter_dl( [H|T], Var ) :-
	append( H, V

*/
debug_segment_lengths( true, Segms ) :-
	maplist( length, Segms, Lens ),
	debug( mlu(k_fold_learn), 'Data fold partition, segment lengths: ~w', [Lens] ).
debug_segment_lengths( false, _Segms ).

k_fold_moduled( Goal, Moduled ) :-
	compare( meta, HasModule, Goal, =(_:_) ),
	k_fold_goal( HasModule, Goal, Moduled ).

k_fold_goal( =, Goal, Goal ).
k_fold_goal( <>, Goal, user:Goal ).



:- suggests( real ).   % move all real components in their own file

k_fold_comparative_statistic_defaults( Defs ) :-
	Defs = [  rerun_ground(true), learner_names(_), names(_) ].

:- multifile( k_fold_comparative_statistic/6 ).

/** k_fold_pairwise_comparisons( +Data, +Learners, +Predictors, -Models, -Statistics, +Opts ).

	Compare M learners, Goals, over N cross sections of Data. 
	On each of the N iterations, the learners are ran on the N-1 sections and tested on
	the hold out Nth section. 

	If a single Predictor is given, (singleton list or unlisted), then it is used
	on all Learerns. Else number should be equal to number of lerners. There
	should be at least two Learners, else you better call k_fold_learn directly
	(see Predictor and Statistic options there).

	Opts
	 * learner_names(Lnames) 
	   used to distinguish the models for each learner, 
	   l<N> is used by default

	 * names(Names)
	   Names for Models. Unlike other non post() options, who are passed
	   on to Post calls via options, this is passed to explicitly. 
	   This not used in main call, and it defaults to model_01...model_NN
	   If a variable, the generated names are returned.

	 * post(Post)
	   Post processing after Models + Statistics are constructed
        Known Post values: 
	   * jitter

	 * statistic_names(StaNames)
	   names for the statistics

	Additional Opts for the Post are allowed. For each Post, all options with matching outer term name
	are stripped and passed. For instance jitter(accuracy(2,'AUC')) is passed to Post jitter, and 
	it signifies that the predictor passes as it is second argument the AUC of the model against the 
	leave out segment. The implementation is expected correct at the predictor end, here we just
	provide means to pass the information to the plotter.

	Also see options for k_fold_learn/4.

	Jitter
	* accuracy(N,Name)
	  the Nth (>1) position of the statistic is accuracy identified by Name

	* accuracy(Pname)
	  predicate name for obtaining the accuracy names, called as call(Pname,N,Name)

	* average_accuracy(Avg=mean)
	  R function for obtaining the single accuracy from the k_fold accuracies

	 * rerun_ground(Rerun=true)
        set to false to avoid re-runing ground models. Convenient for running 
	   comparatives

==
	?- [pack(mlu/examples/stoic)].

	?- stoic.  % use ?- stoic_ng.  % if you don't have library(real).

==


*/
k_fold_pairwise_comparisons( Data, Learners, PredictorsPrv, Models, Statistics, Args ) :-
	options_append( k_fold_comparative_statistic, Args, Opts ),
	k_fold_comparative_statistic_predictors( PredictorsPrv, Learners, Predictors, Opts ),
	options( rerun_ground(Rerun), Opts ),
	k_fold_comparative_statistic_call( Learners, Rerun, 0, Data, Predictors, Models, Statistics, Opts ),
	options( names(Names), Opts ),
	options( learner_names(Lames), Opts ),
	k_fold_posts( Opts, Learners, Models, Posts, Names, Lames, OthOpts ),
	% options( statistic_names(StaNames), Opts ),
	% holds( var(StaNames), VarStaNames ),
	% k_fold_statistic_names_ground( VarStaNames, Predictors, StaNames ),
	% here is where the statistic part happens...
	options( statistic_names(StaNames), Opts ),
	k_fold_comparative_statistic_post( Posts, Models, StaNames, Statistics, Lames, OthOpts ).

/*
k_fold_statistic_names_grouk_fold_comparative_statistic-komos-16.01.04.plnd( false, _Predictors, _StaNames ).
k_fold_statistic_names_ground( true, Predictors, StaNames ) :-
	findall( StaName, (member(P,Predictors),arity(P,StaName,_)), StaNames ).
	*/
	
k_fold_posts( [], Learners, Models, [], Names, Lames, [] ) :-
	holds( var(Names), VarNames ),
	k_fold_posts_names( VarNames, Learners, Models, Names, Lames ).
k_fold_posts( [H|T], Learners, Models, Posts, Names, Lames, Opts ) :-
	holds( H=post(_), IsPost ),
	k_fold_post( IsPost, H, Posts, TPosts, Opts, TOpts  ),
	k_fold_posts( T, Learners, Models, TPosts, Names, Lames, TOpts ).

k_fold_post( true, post(Post), [Post|Posts], Posts, Opts, Opts  ).
k_fold_post( false, NonPost, Posts, Posts, [NonPost|Opts], Opts  ).

k_fold_post_model_names( false, _Models, Lames, Lames ). % for now
k_fold_post_model_names( true, Models, Lames, Names ) :-
	length( Models, NofMs ),
	NofPs is floor( log10(NofMs) ),
	findall( SubNames, (
					 	nth1(N,Models,NModels),
						nth1(N,Lames,Lame),
						% k_fold_z_pad(NModels,Zad),
						findall( Sub, ( nth1(M,NModels,_NMod),
									 k_fold_z_pad(NofPs,M,Zad),
						                at_con([Lame,'_m',Zad],'',Sub)
						              ),  SubNames )
	                   ), Names ).

k_fold_z_pad( NofPs, M, Zad ) :-
	NofMPs is floor( log10(M) ),
	NofZs is NofPs - NofMPs,
	findall( 0, between(1,NofZs,_), Zeros ),
	append( Zeros, [M], Parts ),
	at_con( Parts, '', Zad ).

k_fold_posts_names( true, Learners, Models, Names, Lames ) :-
	holds( ground(Lames), Games ),
	k_fold_posts_lerner_names( Games, Learners, Models, Lames ),
	Models = [Hm|_],
	holds( is_list(Hm), IsList ),
	k_fold_post_model_names( IsList, Models, Lames, Names ).
k_fold_posts_names( false, _Learners, _Models, _Names, _Lames ).

k_fold_posts_lerner_names( true, _Learners, _Models, _Lames ).
k_fold_posts_lerner_names( false, Learners, _Models, Lames ) :-
	sort( Learners, Searners ),
	length( Learners, Llen ),
	length( Searners, Slen ),
	holds( Llen=:=Slen, EnoughNames ),
	k_fold_posts_learner_names_complete( EnoughNames, Learners, Lames ).

k_fold_posts_learner_names_complete( true, Learners, Lames ) :-
	findall( Lame, (member(Learner,Learners),goal_predicate_name(Learner,Lame)), Lames ).
k_fold_posts_learner_names_complete( false, Learners, Lames ) :-
	length( Learners, NofLs ),
	MaxNofZs is floor( log10(NofLs) ),
	% findall( '0', between(1,NofLs,_), Zeroes ),
	findall( Lame, (  
	                  nth1(I,Learners,Learner),
				   NofZs is floor( log10(I) ),
				   Gap is MaxNofZs - NofZs,
				   findall( 0, between(1,Gap,_), Zs ),
				   goal_predicate_name( Learner, Lname ),
				   at_con([Lname,'_',Zs],Lame)
				), 
				      Lames ).

	% atomic_list_concat( Zeroes, Pad ),
	% holds( NofMs =:= 0, Termin ),
	% Pfx = lrn,
	% k_fold_posts_default_names( Termin, 1, NofMs, Pfx, Pad, Lames ).

k_fold_posts_default_names( false, I, NofMs, Pfx, Pad, [Name|Names] ) :-
	atomic_list_concat( [Pfx,'_',Pad,I], '', Name ),
	J is I + 1, 
	holds( J > NofMs, Termin ),
	k_fold_posts_default_names( Termin, J, NofMs, Pad, Names ).
k_fold_posts_default_names( true, _I, _NofMs, _Pad, [] ).

k_fold_comparative_statistic_post( [], _Models, _StaNames, _Statistics, _Lames, _Opts ) :-
	debug( mlu, 'Done with post processing statistics.', [] ).
k_fold_comparative_statistic_post( [Post|Posts], Models, StaNames, Statistics, Lames, Opts ) :-
	( atomic(Post) -> PostName = Post; functor(Post,PostName,_,_) ),
	PostOptWrap =.. [PostName,PostOpt],
	findall( PostOpt, member(PostOptWrap,Opts), PostOpts ),
	k_fold_comparative_statistic_post_call( Post, Models, StaNames, Statistics, Lames, PostOpts ),
	k_fold_comparative_statistic_post( Posts, Models, StaNames, Statistics, Lames, Opts ).

k_fold_comparative_statistic_post_call( cox_cohese, Models, StaNames, Statistics, Lames, _Opts ) :-
	% throw( here(Models,Lames,Opts) ).
	maplist( cox_cohese_summary(StaNames), Models, Statistics, Lames, Files ),
	write( wrote_csv_files(Files) ), nl.

% here: extend to jitter(Which).
% k_fold_comparative_statistic_post_call( jitter, _Models, StaNames, Statistics, Lames, Opts ) :-
k_fold_comparative_statistic_post_call( jitter, Models, StaNames, Statistics, Lames, Opts ) :-
	length( StaNames, Len ),
	k_fold_comparative_statistic_post_call( jitter(Len), Models, StaNames, Statistics, Lames, Opts ).

k_fold_comparative_statistic_post_call( jitter(UpTo), _Models, StaNames, Statistics, Lames, Opts ) :-
	% Stats = [[c(A1p1,A1p2,A1p3),c(A2p1,A2p2,A2p3)],[c(B1p1,B1p2,B1p3),c(B2p1,B2p2,B2p3)]]
	% assume for now inners are always compunds, ...
	Defs = [ onoma(true), accuracy_average(mean), postfix(''), pval(true) ],
	/* */
	% Statistics = [[StaTerm|_],[_|_]],  % for now only supporting pairs...
	% functor( StaTerm, _, Arity ),
	% length( StaNames, Arity ),  % sanity check
	( length(StaNames,UpTo) -> true; throw( fixme(sta_names_of_wrong_length_in_jitter(UpTo,StaNames)) ) ),
	append( Defs, Opts, All ),
	% ExcLast is Arity - 1,
	% ( Lames = ['Age',muts_ose] -> trace; write( no_tracing(Lames) ), nl ),
	k_fold_comparative_statistic_post_call_jitter( UpTo, StaNames, Statistics, Lames, All ).

% this is for single learner, called upstream by maplist/n
%
cox_cohese_summary( StaNames, Models, StatisticsTerm, Lame, File ) :-
	findall( Vars, (member(Mod,Models), VarS <- rownames( summary(Mod)$coef ),en_list(VarS,Vars)), VarsNest ),
	findall( CoeM, (member(Mod,Models), CoeM <- summary(Mod)$coef), CoeMNest ),
	flatten( VarsNest, FlatVars ),
	sort( FlatVars, AllVars ),
	findall( row(Var,Appr,Coef,Z,Pv), 
				( member(Var,AllVars),
				  findall( 1, (member(VarsA,VarsNest),memberchk(Var,VarsA)), Ones ),
				  sumlist( Ones, Appr ),
				  findall( c(Coef,Z,Pv), ( nth1(VarsAP,VarsNest,VarsA), nth1(VarP,VarsA,Var),
							            nth1(VarsAP,CoeMNest,CoemA), nth1(VarP,CoemA,Coem),
							            Coem = [Coef,_,_,Z,Pv]
				                  ),
							   		CoefTerms
							      ),
				  maplist( arg(1), CoefTerms, Coefs ), Coef <- median(Coefs),
				  maplist( arg(2), CoefTerms, Zs ),    Z    <- median(Zs),
	 			  maplist( arg(3), CoefTerms, Pvs ),   Pv   <- median(Pvs)
				),
			 Varrows
			),
	% StaNamesT =.. [row|StaNames],
	% StaMtx    = [StaNamesT|Statistics],
	% Statistics = [StatisticsA|_],
	% StatisticsA = [StatsATerm|_],
	% functor( StatsATerm, _, Arity ),
	% N is Arity - UpTo,
	maplist( remove_last_n_args(1), StatisticsTerm, StatisticsOnly ),
	rows_cohese( Varrows, StatisticsOnly, 5, 4, Csv ),
	atomic_list_concat( [cox,appears,summary,Lame], '_', Stem ),
	os_ext( csv, Stem, File ),
	write( writing_on_file(File) ), nl,
	once( append(StaNamesOnly,[_],StaNames) ),
	Hdr =.. [row,var,appears,coef,z,pval|StaNamesOnly],
	csv_write_file( File, [Hdr|Csv] ).
	% csv_write_file( [row('var',coef,z,pval,)|Csv], File, [match_arity(false)] ).
	
	/*
remove_last_arg_list( Terms, Reds ) :-
	maplist( remove_last_arg, Terms, Reds ).
	*/

remove_last_n_args( N, Term, Red ) :-
	Term =.. [Name|Args],
	length( List, N ),
	once( append(Goods,List,Args) ),
	Red =.. [Name|Goods].

rows_cohese( [], [], _N, _M, [] ) :- !.
rows_cohese( [], Posts, N, _M, Rs ) :- !,
	findall( '', between(1,N,_), Elist ),
	Eterm =.. [row|Elist],
	findall( R, (member(P,Posts),arg_append(Eterm,P,R)), Rs ).
rows_cohese( Pres, [], _N, M, Rs ) :- !,
	findall( '', between(1,M,_), Elist ),
	Eterm =.. [row|Elist],
	findall( R, (member(P,Pres),arg_append(P,Eterm,R)), Rs ).
rows_cohese( [Pre|Pres], [Post|Posts], N, M, [R|Rs] ) :-
	arg_append( Pre, Post, R ),
	rows_cohese( Pres, Posts, N, M, Rs ).

/* the following is a "wrong" stub
	 if we jitter a multi predictor output, 
	 its term arguments should 
	 1st  
*/
k_fold_comparative_statistic_post_call_jitter( 0, _StaNames, _Stats, _LNames, _Opts ) :- !.
k_fold_comparative_statistic_post_call_jitter( 1, StaNames, [StsA,StsB], [LA,LB], Opts ) :- !,
	maplist( arg(1), StsA, NtsA ),
	maplist( arg(1), StsB, NtsB ),
	% nth1( 1, StaNames, NStaName ),
	StaNames = [NStaName],
	jitter_plot( NtsA, NtsB, LA, LB, NStaName, [], Opts ).
k_fold_comparative_statistic_post_call_jitter( N, StaNames, [StsA,StsB], [LrnA,LrnB], Opts ) :-
	maplist( arg(1), StsA, NtsA ),
	maplist( arg(1), StsB, NtsB ),
	StaNames = [StName|_],
	options( accuracy_average(AccFunc), Opts ),
	holds( 2>N, Termin ),
	k_fold_comparative_statistic_jitter_accuracies( Termin, 2, N, StsA, StsB, StaNames, AccFunc, AccPairs ),
	( jitter_plot( NtsA, NtsB, LrnA, LrnB, StName, AccPairs, Opts ) -> true
			; write( jitter_plot_failure(NtsA,NtsB,LrnA,LrnB,StName,AccPairs,Opts) ), nl ).

k_fold_comparative_statistic_jitter_accuracies( true, _O, _N, _StsA, _StsB, _StaNames, _AccFunc, [] ).
k_fold_comparative_statistic_jitter_accuracies( false, L, N, StsA, StsB, SttNames, AccFunc, [SttName=c(SttAvgA,SttAvgB)|TPairs] ) :-
	maplist( arg(L), StsA, NtsA ),
	maplist( arg(L), StsB, NtsB ),
	nth1( L, SttNames, SttName ),
	AvgARall =.. [AccFunc,NtsA],
	AvgBRall =.. [AccFunc,NtsB],
	SttAvgA <- AvgARall,
	SttAvgB <- AvgBRall,
	M is L + 1,
	holds( M > N, Termin ),
	k_fold_comparative_statistic_jitter_accuracies( Termin, M, N, StsA, StsB, SttNames, AccFunc, TPairs ).
	

jitter_plot( NtsA, NtsB, LrnA, LrnB, StName, AccPairs, Opts ) :- % fixme, fish odir() here
	write( acc_pairs(AccPairs) ), nl,
	length( NtsA, Len ),
	length( NtsB, Len ),
	options( onoma(Onoma), Opts ),
	% fixme throw error if not == 
	at_con( [StName,LrnA,vs,LrnB], '_', Base ),
	options( postfix(Psf), Opts ),
	os_ext( pdf, Base, PdfFPrv ),
	os_postfix( Psf, PdfFPrv, PdfF ),
	append( NtsA, NtsB, Nts ),
	% <- plot(jitter(rep(1:2,each=Len)), Nts, ylim=c(0.5,1), axes='FALSE', ylab="Harrel's C", xlab="", pch=20, asp=3)
	( Onoma == true -> maplist(onoma,[StName,LrnA,LrnB],[Ylab,OnmA,OnmB])
	                   ; 
				    Ylab=StName, OnmA=LrnA, OnmB=LrnB 
	),
	( AccPairs = [_|_] -> Xlim=c(0.5,3),Width=9; Xlim=c(1,2), Width=7 ),
	<- pdf( +PdfF, width=Width, height=7 ),
	<- plot(jitter(rep(1:2,each=Len)), Nts, ylim=c(0.5,1), xlim=Xlim, axes='FALSE', ylab=+Ylab, xlab="", pch=20, asp=3),
	% <- axis(side=1, at=1:2, labels = c(+OnmA,+OnmB), col.axis=c("red","green") ), % col.lab=c("#4B5320","#21ABCD") ),
	ColA = "#177245", % dark spring green
	ColB = "#21ABCD", % ball blue
	<- axis(side=1, at=1, labels = c(+OnmA), col.axis=ColA ),
	% <- axis(side=1, at=1, labels = c(+OnmA), col.axis=c("#03C03C") ), % 
	% <- axis(side=1, at=1, labels = c(+OnmA), col.axis=c("#006B3C") ), % 
	% <- axis(side=1, at=1, labels = c(+OnmA), col.axis=c("#21CDAB") ), % 
	<- axis(side=1, at=2, labels = c(+OnmB), col.axis=ColB ),
	<- axis(side=1, at=1:2, labels= c("","") ), % col.lab=c("#4B5320","#21ABCD") ),
	<- axis(side=2, at=(5:10)/10),
	at_con( [Ylab,for,LrnA,vs,LrnB,models], ' ', Main ),
	% <- title("Harrel's C for baseline vs full model"),
	<- title( +Main ),
	MeanA <- mean(NtsA),
	MeanB <- mean(NtsB),
	Hw is 0.1, MinXA is 1 - Hw, MaxXA is 1+Hw, MinXB is 2 - Hw, MaxXB is 2 + Hw,
	<- segments( c(MinXA,MinXB), c(MeanA,MeanB), c(MaxXA,MaxXB), c(MeanA,MeanB) ),
	( options(pval(true),Opts) -> 
		Ttest <- t.test( NtsA, NtsB, paired='TRUE' ),
		memberchk( p.value=TtPval, Ttest ),
		atom_number( TtPvalAt, TtPval ),
		once( sub_atom(TtPvalAt,0,6,_,TtPvalAtPfx) ),
		<- text( 1.5, 1.0, +TtPvalAtPfx, col="#E25822" ) % flame
		;
		true
	),
	% <- text( 1.5, 1.0, +TtPvalAtPfx, col="#87A96B" ),
	% <- text( 1.5, 1.0, +TtPvalAtPfx, col="#CC4E5C" ),
	% <- text( 1.5, 1.0, +TtPvalAtPfx, col="#E9692C" ), % deep carrot orange

	% flighing low...
	jitter_plot_averaged_statistics( AccPairs, 1, 0.03, ColA, ColB ),
	<- dev.off(),
	debug_call( mlu(k_fold_comparative_statistic), wrote, PdfF ).

jitter_plot_averaged_statistics( [], _Y, _Step, _ColA, _ColB ).
jitter_plot_averaged_statistics( [SttName=c(SttA,SttB)|T], Y, Step, ColA, ColB ) :-
	<- text( 3, Y, +SttName, pos=2 ),
	maplist( trunc(6), [SttA,SttB], [TruA,TruB] ),
	<- text( 3, Y-Step, TruA, col=ColA, pos=2 ),
	<- text( 3, Y-(2*Step), TruB, col=ColB, pos=2 ),
	Z is Y - (4*Step),
	jitter_plot_averaged_statistics( T, Z, Step, ColA, ColB ).

k_fold_comparative_statistic_call( [], _ReR, _L, _Data, [], [], [], _Opts ).
k_fold_comparative_statistic_call( [LGoal|Learners], ReR, L, Data, [PGoal|Prcts], [LMs|Ms], [LSts|Sts], Opts ) :-
	holds( (ReR==false,ground(LMs)), DontRun ),
	Self = mlu(k_fold_comparative_statistic),
	k_fold_comparative_statistic_call_rerun( DontRun, LGoal, L, Data, PGoal, Self, LMs, LSts, Opts ),
	% en_list( LearnedMs, LearnedMsL ),
	% en_list( LearnedSts, LearnedStsL ),
	% k_fold_comparative_statistic_stitch( LearnedMsL, LearnedStsL, Ms, Sts, TMs, TSts ),
	M is L + 1,
	k_fold_comparative_statistic_call( Learners, ReR, M, Data, Prcts, Ms, Sts, Opts ).

k_fold_comparative_statistic_call_rerun( true, LG, _L, _D, _P, Self, Ms, _Sts, _Opts ) :-
	Mess = 'Skipping k_fold_learn for learner:~w with ground Models: ~w',
	debug( Self, Mess, [LG,Ms] ).
k_fold_comparative_statistic_call_rerun( false, LGoal, L, Data, PGoal, Self, Ms, Sts, Opts ) :-
	k_fold_comparative_model_stem( LGoal, L, LPstem, Opts ),
	KFOpts = [model_name_stem(LPstem),predictor(PGoal),statistics(Sts)|Opts],
	debug( Self, 'Calling k_fold_learn for learner: ~w', LGoal ),
	k_fold_learn( Data, LGoal, Ms, KFOpts ),
	debug_call( Self, length, models_learnt/Ms ).

k_fold_comparative_model_stem( _LGoal, L, LPstem, Opts ) :-
	memberchk( learner_names(Lnames), Opts ),
	nth1( L, Lnames, Lname ),
	ground( Lname ),
	!,
	downcase_atom( Lname, LPstem ). % fixme, also replace ' ' by _
k_fold_comparative_model_stem( LGoal, L, LPstem, _Opts ) :-
	goal_predicate_name( LGoal, LPname ),
	% shall we also check for groundness of LM here ?
	LetterCode is 0'A + L,
	atom_codes( LetterAtom, [LetterCode] ),
	downcase_atom( LPname, LPdown ),
	at_con( [LPdown,LetterAtom], '_', LPstem ).

k_fold_comparative_statistic_stitch( [], [], TMs, TSts, TMs, TSts ).
k_fold_comparative_statistic_stitch( [M|Ms], [St|Sts], [M|MMs], [St|MSts], TMs, TSts ) :-
	k_fold_comparative_statistic_stitch( Ms, Sts, MMs, MSts, TMs, TSts ).

k_fold_comparative_statistic_predictors( PredictorsPrv, Learners, Predictors, _Opts ) :-
	en_list( PredictorsPrv, PredictorsL ),
	length( PredictorsL, Plen ),
	length( Learners, Llen ),
	holds( Plen=:=1, Singleton ),
	k_fold_comparative_statistic_predictors_complement( Singleton, Plen, Llen, PredictorsL, Predictors ).

k_fold_comparative_statistic_predictors_complement( true, 1, Llen, [PredictorInv], Predictors ) :-
	findall( Predictor, (between(1,Llen,_),copy_term(PredictorInv,Predictor)), Predictors ).
k_fold_comparative_statistic_predictors_complement( false, Plen, Llen, Predictors, Predictors ) :-
	holds( Plen=:=Llen, EqLen ),
	k_fold_comparative_statistic_predictors_complement_equal_length( EqLen, Plen, Llen ),

k_fold_comparative_statistic_predictors_complement_equal_length( true, _Plen, _Llen ).
k_fold_comparative_statistic_predictors_complement_equal_length( false, Plen,  Llen ) :-
	% throw( pack_error(mlu,true,,learners_predictors_unequal_length(Llen,Plen)) ).
	Pred = k_fold_comparative_statistic/6,
	throw( pack_error(mlu,Pred,lengths_mismatch(learners,predictors,Llen,Plen)) ).



k_fold_pairwise_predictions_defaults( Args, Defs ) :-
	memberchk( learners(Lrns), Args ),
	( memberchk(pairs(_),Args) -> TDefs = []; 
		% kv_list( Lrns, Pairs ),
		TDefs = [pairs(_Pairs)]
	),
	( memberchk(learner_names(_Lames),Args) ->
		TDefs = Defs
		;
		Defs = [learner_names(Lrns)|TDefs]
	).

/** k_fold_pairwise_predictions( +Dat, +Learners, +Predictor, -Models, -Stats, +Opts ).

	Run k_fold_pair_predictions/7 on pairs of Learners on a single k_fold segmentation. 
	By default all pairwise comparisons are considered.

	Opts
	* pairs(Pairs)
	  list or single pair (L1-L2) of predictions to consider

	@see options for k_fold_pair_predictions/7
	@tbd allow distinct Predictors

*/
k_fold_pairwise_predictions( Dat, Lrs, PrG, Mls, Sts, Args ) :-
	Self = k_fold_pairwise_predictions,
	options_append( k_fold_pairwise_predictions, Args, Opts, extra_arg(learners(Lrs)) ),
	% here( Dat, Lrs, Prs, Mls, Sts ).
	k_fold_learn_list( Lrs, 1, Self, Dat, PrG, _Segs, Mls, Sts, Opts ),
	debug( k_fold_pairwise_predictions, 'done learning', [true] ),
	% k_fold_pairs_predictions( Pairs, Dat, Lrs, Mls, 
	options( learner_names(Lames), Opts ),
	options( pairs(Pairs), Opts ),
	k_fold_pairwise_posts_collect( Opts, Posts, OthOpts ),
	findall( Post-PostOpts, ( member(Post,Posts),term_name(Post,PostName),PostTerm=..[PostName,PostOpt],
	                          findall(PostOpt,member(PostTerm,OthOpts),PostOpts) ), PostPairs ),
	options( statistic_names(StaNames), Opts ),
	maplist( k_fold_postfix_multi_learner, Lrs, MrsNest ),
	flatten( MrsNest, Mrs ),
	k_fold_pairwise_learners_predictions( Mrs, Pairs, Lames, StaNames, Mls, Sts, PostPairs ).

k_fold_postfix_multi_learner( multi(GPsfxs,Goal), Posted ) :-
	!,
	findall( Postfixed, ( member(Psfx,GPsfxs), goal_predicate_name(Goal,Stem),
					  atomic_list_concat( [Stem,Psfx], '_', Postfixed )
	                    ), Posted ).
k_fold_postfix_multi_learner( Posted, Posted ).

k_fold_pairwise_learners_predictions( [], _Pairs, [], _StaNames, [], [], _PostPairs ).
k_fold_pairwise_learners_predictions( [Lr|Lrs], Pairs, [Lm|Lms], StaNames, [Ml|Mls], [St|Sts], PostPairs ) :-
	k_fold_pairwise_predictions_nest( Lrs, Lms, StaNames, Pairs, Lr, Lm, Mls, Ml, Sts, St, PostPairs ),
	k_fold_pairwise_learners_predictions( Lrs, Pairs, Lms, StaNames, Mls, Sts, PostPairs ).

k_fold_pairwise_predictions_nest( [], [], _StaNames, _Pairs, _Lr1, _Lm, [], _Ml1, [], _St1, _PostPairs ).
k_fold_pairwise_predictions_nest( [Lr2|Lrs], [Lm2|Lms], StaNames, Pairs, Lr1, Lm1, [Mls2|TMls], Mls1, [Sts2|TSts], Sts1, PostPairs ) :-
	% options( names(Names), Opts ),
	holds( memberchk(Lr1-Lr2,Pairs), PairFires ),
	k_fold_pairwise_pair_posts( PairFires, Lm1, Lm2, Mls1, Mls2, StaNames, Sts1, Sts2, PostPairs ),
	k_fold_pairwise_predictions_nest( Lrs, Lms, StaNames, Pairs, Lr1, Lm1, TMls, Mls1, TSts, Sts1, PostPairs ).

k_fold_pairwise_pair_posts( true, Lm1, Lm2, Mls1, Mls2, StaNames, Sts1, Sts2, PostPairs ) :-
	write( lm1(Lm1)-lm2(Lm2) ), nl,
	k_fold_pairwise_posts( PostPairs, Lm1, Lm2, Mls1, Mls2, StaNames, Sts1, Sts2 ).
k_fold_pairwise_pair_posts( false, _Lm1, _Lm2, _Mls1, _Mls2, _StaNames, _Sts1, _Sts2, _PostPairs ).

k_fold_pairwise_posts( [], _Lm1, _Lm2, _Mls, _Mls2, _StaNames, _Sts1, _Sts2 ).
k_fold_pairwise_posts( [Post-PostOpts|Posts], Lm1, Lm2, Mls1, Mls2, StaNames, Sts1, Sts2 ) :-
	% findall( PostOpt, member(PostTerm,Opts), PostOpts ),
	trace,
	write( doing( Post, PostOpts, Lm1, Lm2, Mls1, Mls2, StaNames, Sts1,Sts2) ), nl,
	k_fold_comparative_statistic_post_call( Post, [Mls1,Mls2], StaNames, [Sts1,Sts2], [Lm1,Lm2], PostOpts ),
	k_fold_pairwise_posts( Posts, Lm1, Lm2, Mls1, Mls2, StaNames, Sts1, Sts2 ).

k_fold_pairwise_posts_collect( [], [], [] ).
k_fold_pairwise_posts_collect( [H|T], Posts, Other ) :-
	holds( H=post(Post), IsPost ),
	k_fold_pairwise_is_post( IsPost, Post, H, Posts, Other, TPosts, TOther ),
	k_fold_pairwise_posts_collect( T, TPosts, TOther ).

k_fold_pairwise_is_post( true, Post, _H, [Post|Posts], Other, Posts, Other ).
k_fold_pairwise_is_post( false, _Post, H, Posts, [H|Other], Posts, Other ).

/*
	Copts = [ include_stem(model), learner_names(['Age','Clinical']),
			% statistic_names([prediction,auc,harrel,r_square]), 
			segments(Segms),
			rerun_ground(false)
		    ],
	Learners1 = [surv_clm('Agegrp5'),clinical],
	k_fold_comparative_statistic( Data, Learners1, auc(365), Models1, Stats1, Copts ),
	Models1 = [AgeMls,_],
	Stats1  = [AgeSts,_],
	debug( lasso, 'Age x clinical Models: ~w, Stats: ~w', [Models1,Stats1] ),
	% here, Segms is not ground...
	Mopts = [ include_header(true), post(jitter), 
	          include_stem(model), learner_names(['Age','Lasso_min']),
			statistic_names([prediction,auc,harrel,r_square]), segments(Segms),
			rerun_ground(false)
		    ],
	Learners2 = [surv_clm('Agegrp5'),lasso_min],
	k_fold_comparative_statistic( Data, Learners2, auc(365), [AgeMls,LminMls], [AgeSts,LminSts], Mopts ),
	debug( lasso, 'Age x lasso models: ~w, Stats: ~w', [[AgeMls,LminMls],[AgeSts,LminSts]] ).
	*/

k_fold_learn_list( [], _N, _Self, _Dat, _PrG, _Segs, [], [], _Opts ).
% k_fold_learn_list( [LGoalPrv|LGoals], L, Self, Data, PrG, Segs, [Mls|TMls], [Sts|TSts], Opts ) :-
k_fold_learn_list( [LGoalPrv|LGoals], L, Self, Data, PrG, Segs, AMls, ASts, Opts ) :-
	( LGoalPrv = multi(Msfxs,LGoal) -> length(Msfxs,Multi), Add is Multi ; LGoal = LGoalPrv, Multi is 0, Add is 1 ),
	k_fold_comparative_model_stem( LGoal, L, LPstem, Opts ),
	KFOpts = [segments(Segs),model_name_stem(LPstem),predictor(PrG),statistics(Sts)|Opts],
	debug( Self, 'Calling k_fold_learn for ~d-th learner: ~w', [L,LGoal] ),
	k_fold_learn( Data, LGoalPrv, Mls, KFOpts ),
	debug_call( Self, length, models_learnt/Mls ),
	k_fold_learn_list_multi_learned( Multi, Mls, Sts, AMls, ASts, TMls, TSts ),
	M is L + Add,
	k_fold_learn_list( LGoals, M, Self, Data, PrG, Segs, TMls, TSts, Opts ).

k_fold_learn_list_multi_learned( 0, Mls, Sts, [Mls|TMls], [Sts|TSts], TMls, TSts ) :- !.
k_fold_learn_list_multi_learned( M, Mls, Sts, AMls, ASts, TMls, TSts ) :-
	Mls = [HMls|_],
	length( HMls, HLen ),
	( HLen =:= M -> true; throw( multi_learned_mis_match(HMls) ) ),
	findall( IMls, (between(1,M,I),findall(IMl,(member(FMls,Mls),nth1(I,FMls,IMl)),IMls)), NewMls ),
	findall( ISts, (between(1,M,I),findall(ISt,(member(FSts,Sts),nth1(I,FSts,ISt)),ISts)), NewSts ),
	append( NewMls, TMls, AMls ),
	append( NewSts, TSts, ASts ).

	% k_fold_learn_list_multi_learned_m( Mls, M, Sts, AMls, ASts, TMls, TSts ).

k_fold_learn_list_multi_learned_m( [], 0, [], AMls, ASts, AMls, ASts ) .
k_fold_learn_list_multi_learned_m( [Mls|RMls], M, [Sts|RSts], [Mls|MMls], [Sts|MSts], TMls, TSts ) :-
	L is M - 1,
	k_fold_learn_list_multi_learned_m( RMls, L, RSts, MMls, MSts, TMls, TSts ).


/** mlu_errors.

Documentation predicate.

Pack mlu uses pack pack_errors for throwing errors. 

File mlu_errors.pl defines local errors within the pack_errors infrastructure.
*/

mlu_errors.

:- multifile( pack_errors:message/3 ).

pack_errors:message( fold_data_insufficient(Dlen,N) ) -->
	['Insufficient length of data (~d) as ~d folds are required'-[Dlen,N]].
pack_errors:message( fold_data_residual(Dlen) ) -->
	['Residual data of length: ~d while splitting folds'-[Dlen]].
pack_errors:message( learners_predictors_unequal_length(Llen,Plen) ) -->
	['Learners and predictors have different lengths: ~d and ~d respectively'-[Llen,Plen]].

/*
pack_errors:message( missing_alias(Os) ) -->
	['OS entity: ~w, looks like aliased but alias does not exist.'-[Os]].

pack_errors:message( os_postfix_lists(Sc,Fr) ) -->
	['os_postfix/4: only one of args 2 and 4 should be a list (options).\nFound,\n2nd: ~w\n4th:~w'-[Sc,Fr]].
	*/


:- lib( options ).


n_breaks_defaults( Defs ) :-
	Defs = [centre(false),fixed_width(false)].

/** n_breaks( +Vector, +N, -Breaks, -Opts ).

For a vector of values, create N break points.

The number of Breaks is always odd when Centre is true. This interprets odd N as the number of 
break points, even N it is taken to be the number of 
intervals.

==
?- n_breaks( [1,3,4,4,5,5,6,8], 4, Bs, [] ).
Bs = [1.0, 2.75, 4.5, 6.25, 8.0].

?- n_breaks( [0.21,3,4,4,5,5,6,8], 4, Bs, [centre(1)] ).
Bs = [0.21, 0.4075, 0.605, 0.8025, 1.0, 2.75, 4.5, 6.25, 8.0].

?- n_breaks( [0.21,3,4,4,5,5,6,8], 4, Bs, [centre(1),fixed_width(true)] ).
Bs = [-6.0, -4.25, -2.5, -0.75, 1.0, 2.75, 4.5, 6.25, 8.0].



==

Opts 
    
  * centre(Centre=false)
    when an arithmetic value is given, the breaks are symmetrically split left and right of Centre

  * fixed_width(Sym=false)
    if true and Centre arithmetic, the shorter of the left or right is extended to keep
    the breaks of fixed width

@author  nicos angelopoulos
@version 0.1 2015/5/27

@tbd add some polymorphism for Vector

*/

n_breaks( Vals, N, Breaks, Args ) :-
	options_append( n_breaks, Args, Opts ),
	options( centre(Centre), Opts ),
	NumN is N,
	n_breaks_centre( Centre, Vals, NumN, Breaks, Opts ).

n_breaks_centre( Centre, Vals, N, Breaks, Opts ) :-
	number(Centre),
	!,
	options( fixed_width(Fixed), Opts ),
	n_breaks_fixed_centre( Fixed, Centre, Vals, N, Breaks ).

n_breaks_centre( _Centre, Vals, N, Breaks, _Opts ) :-
	max_min_list( Vals, Max, Min ),
	n_breaks_range( Min, Max, N, Breaks ).

n_breaks_fixed_centre( true, Centre, Vals, N, Breaks ) :-
	max_min_list( Vals, Max, Min ),
	RelMax is max( abs(Max-Centre), abs(Centre - abs(Min)) ),
	ConcMax is Centre + RelMax,
	ConcMin is Centre - RelMax,
	Half is floor( N ),
	n_breaks_range( ConcMin, Centre, Half, Left ),
	n_breaks_range( Centre, ConcMax, Half, [_|Right] ),
	append( Left, Right, Breaks ).
n_breaks_fixed_centre( false, Centre, Vals, N, Breaks ) :-
	max_min_list( Vals, Max, Min ),
	Half is floor( N ),
	n_breaks_range( Min, Centre, Half, Left ),
	n_breaks_range( Centre, Max, Half, [_|Right] ),
	append( Left, Right, Breaks ).

n_breaks_range( Min, Max, N, Breaks ) :-
	Ivl is abs(Max-Min) / N,
	findall( Break, (between(0,N,I),Break is Min + (Ivl * I)), Breaks ).



%% compare( +Type, ?Op, +Term1, +Term2 ).
%
% Common interface for compare/3 and compare_arithmetic/3, which also
% allows for meta calls. In this case Op is = iff call on 
% call( Term2, Term1 ) succeeds, else it is <>.
% Type should be one of =|meta|=, =|term|= or =|arithmetic|= respectively.
%
% >:< is a special Op, that is always truw (under all interfaces)
% 
%==
% ?- compare( term, Op, 3, 3.0 ).
% ?- compare( arithmetic, Op, 3, 3.0 ).
% ?- compare( meta, Op, 3, =(3.0) ).
% Op = <> .
% ?- compare( meta, Op, 3, =:=(3.0)).
% Op =  (=).
% ?- compare( meta, Op, 3, =(3.0) ).
% Op = <> .
% 
% ?- compare( meta, Op, 3, =:=(3.0) ).
% Op =  (=).
% ?- compare( term, >:<, 3, 2 ).
% ?- compare( arithmetic, >:<, 3, 2 ).
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/2/16
% @version  0.2 2016/2/17, added special operator >:< 
%
compare( _, Op, _Term1, _Term2 ) :-
	ground( Op ),
	Op = >:<,
	!.

compare( term, Op, Term1, Term2 ) :-
	compare( Op, Term1, Term2 ).
compare( arithmetic, Op, Term1, Term2 ) :-
	compare_arithmetic( Op, Term1, Term2 ).
compare( meta, Op, Term1, Term2 ) :-
	( call(Term2,Term1) ->
		Op = (=)
		;
		Op = (<>)
	).

/** holds( +Goal, -Holds ).
    
    Goal is called deterministacally with =|Holds = true|= iff Goal
    succeeds. Else, =|Holds = false|=.

    Note that if Holds is instantiated, Goal will still be called.

==
?- holds( X=3, Holds ).
X = 3,
Holds = true.

?- holds( 4=3, Holds ).
Holds = false.

?- holds( member(X,[a,b]), Holds ).
X = a,
Holds = true.


?- holds( member(X,[a,b]), non_true ).
false.

?- holds( (write(x),nl), non_true ).
x
false.
==

@author nicos angelopoulos
@version  0.1 2015/12/9

*/
holds( Goal, Holds ) :-
	call( Goal ),
	!,
	Holds = true.
holds( _Goal, false ).

/** n_digits_min( +N, +Number, -Padded ).

Padded is the atom coresponding to Number with the possible
addition of leading 0s to pad the length to a minimum of legth = N.

==
 ?- n_digit_integer_min( 2, 2, Atom ).
 Atom = '02'.
==

	@see n_digits/3 for a procrustean version

*/

n_digits_min( N, Int, Padded ) :-
	number_codes( Int, Codes ),
	length( Codes, Len ),
	Pad is N - Len,
	findall( 0'0, between(1,Pad,_), Zs ),
	append( Zs, Codes, PaddedCs ),
	atom_codes( Padded, PaddedCs ).



goal_predicate_name( Var, _ ) :-
	var(Var),
	!,
	fail. % throw().
goal_predicate_name( _Mod:Goal, Pname ) :-
	!,
	arity( Goal, Pname, _ ).
goal_predicate_name( Goal, Pname ) :-
	arity( Goal, Pname, _ ).

%% at_con( ?List, +Sep, ?Atom ).
% 
% 	Mostly a nickname for atomic_list_concat/3  because this is 
%    a well used predicate and the system name is too long, but also
%
%   * allows 
% 	for operators in Sep and a variable Sep get instantiated to '_'. 
%    (Some operators are now allowed in main atomic_list_concat/3, 
%     but this was not the case before).
% 
%   * if List contains '' that is skipped before calling the atomic_list_concat/3.
%   
%   * Sep = ''  is allowed in -List +Atom modality
% 
% ==
% ?- at_con( [a,b,c], _, Abc ).
% Abc = a_b_c.
% ?- at_con( [a,b,'',c], -, Abc ).
% Abc = 'a-b-c'.
% ?- at_con( Parts, '', abc ).
% Parts = [a, b, c].
% ==
%
% @author nicos angelopoulos
% @version  0.2 2014/7/15  added avoidance of ''
%

at_con( List, Atom ) :-
	at_con( List, '', Atom ).

at_con( ListPrv, SepPrv, Atom ) :-
	at_con_separator( SepPrv, Sep ),
	at_con_list( ListPrv, List ),
	seperator_at_con( Sep, List, Atom ).

seperator_at_con( '', List, Atom ) :-
	var( List ),
	!,
	findall( Sub, sub_atom(Atom,_,1,_,Sub), List ).
seperator_at_con( Sep, List, Atom ) :-
	atomic_list_concat( List, Sep, Atom ).

at_con_list( Prv, List ) :-
	% var( Prv ),
	\+ ground( Prv ),
	!,
	List = Prv.
at_con_list( [], [] ).
at_con_list( [H|T], List ) :-
	at_con_head_list( H, List, Tail ),
	at_con_list( T, Tail ).

at_con_head_list( '', List, List ) :- !.
at_con_head_list( Elem, [Elem|List], List ).

at_con_separator( Provisional, Sep ) :-
	var( Provisional ),
	!,
	Sep = '_'.
at_con_separator( Provisional, Sep ) :-
	( current_op(_,_,xfx);current_op(_,_,yfx) ),
	!,
	term_to_atom( Provisional, Sep ).
at_con_separator( Sep, Sep ) :- !.

%% arity( ?Term, ?Name, ?Arity ).
%
%  This is the permissive version, if we detect atomic
%  we use functor/3 (the old stuff), otherwise we call 
%  compound_name_arity/3.
%  
arity( Term, Name, Arity ) :-
	current_predicate( compound_name_arity/3 ),
	\+ atomic( Term ),
	!,
	compound_name_arity( Term, Name, Arity ).
arity( Term, Name, Arity ) :-
	functor( Term, Name, Arity ).

/** onoma( Atom, Onoma ).

	As it is my name day today, this is a Greekified version
	of upcase_first/1.

	@author nicos angelopoulos
	@version  0.2 2015/12/6 St Nicholas day (Greek calendar)

*/
onoma( Atom, Onoma ) :-
	sub_atom( Atom, 0, 1, N, First ),
	upcase_atom( First, Capital ),
	sub_atom( Atom, 1, N, 0, Rest ),
	atom_concat( Capital, Rest, Onoma ).

%% trunc( +P, +Num, -Trunc ).
%
%  Arg rehash of float_truncate/3, with a shorter name too.
trunc( P, Num, Trunc ) :-
	float_truncate( Num, P, Trunc ).

%% float_truncate( +Num, +P, -Trunc ).
%
% Truncate a float to a maximum of P decimals.
%
% Wrote this to cope with R-3.1.0's inability to read long floats.
%
%==
% ?- L is 0.9344889482650001, assert( l(L) ).
% ?- l(L), float_truncate( L, 10, L10 ).
% L10 = 0.9344889483
%
% ?- l(L), float_truncate( L, 20, L20 ).
% L = L20, L20 = 0.9344889482650001.
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/10/28
%
float_truncate( Num, P, Trunc ) :-
 	TenP is 10 ^ P,
	Pow is Num * TenP,
	float_truncate_powered( Pow, Num, TenP, Trunc ).

float_truncate_powered( Pow, Num, _TenP, Trunc ) :-
	Fpow is float_fractional_part( Pow ), 
	0.0 =:= Fpow,     % then we need no truncation, leave the input alone
	!,                % otherwise we might changed it due to floating point
	Trunc is Num.     % arithmetics limitions.
float_truncate_powered( Pow, _Num, TenP, Trunc ) :-
	Trunc is integer(Pow)/TenP.



/** arg_append( +Term, +App, -NewTerm ).

	Append new arguments to a term or atom. App can be a term, an atomic or a list.

==
arg_append( x(a,b), w(y,z), New ).
New = x(a, b, y, z).

arg_append( x(a,b), [y,z], New ).
New = x(a, b, y, z).

arg_append( x(a,b), y, New ).
New = x(a, b, y).

==

 @author nicos angelopoulos
 @version  0.2 2014/03/28, brought test for list first, as it is also a compound!
 @version  0.3 2015/01/12, allow Term to be an atom

*/
arg_append( Term, AppList, New ) :-
	is_list( AppList ),
	!,
	( compound(Term,Tname,TArgs) ->
		true
		;
		atom(Term), % fixme: atomic? 
		Tname = Term,
		TArgs = []
	),
	% Term =.. [Tname|TArgs],
	append( TArgs, AppList, NArgs ),
	compound( New, Tname, NArgs ).
	% New =.. [Tname|NArgs].
arg_append( Term, AppTerm, New ) :-
	compound( AppTerm ), 
	!,
	% AppTerm =.. [_ATname|ATArgs],
	% Term =.. [Tname|TArgs],
	compound( AppTerm, _ATname, ATArgs ),
	compound( Term, Tname, TArgs ),
	append( TArgs, ATArgs, NArgs ),
	% New =.. [Tname|NArgs].
	compound( New, Tname, NArgs ).
arg_append( Term, AppAtomic, New ) :-
	atomic( AppAtomic ),
	% Term =.. [Tname|TArgs],
	compound( Term, Tname, TArgs ),
	append( TArgs, [AppAtomic], NArgs ),
	% New =.. [Tname|NArgs].
	compound( New, Tname, NArgs ).


/** kv_list( +List, -KVs ).

	Create all the reflective K-V pairs possible from List.
Each pair of K-V, V-K only appears once as K-V where K comes before V in List.
This excludes treatment of duplicates, which are considered
unique elements.

==
 ?- kv_list( [a,b,c], KVs ).
 KVs = [a-b, a-c, b-c].
 
 ?- kv_list( [a,b,c,a], KVs ).
 KVs = [a-b, a-c, a-a, b-c, b-a, c-a].
==

@author nicos angelopoulos
@version  0.1 2015/12/23

*/
kv_list( List, Kvs ) :-
	findall( Head-Elem, (tail_list(List,[Head|Tail]),member(Elem,Tail)), Kvs ).



/** term_name( +Term, -Name ).

True iff Name is an atom representation of the name of Term defined as:

  * Term
    when Term is an atom

  * Atomed
    with atom_codes(Term,Codes),atom_codes(Atomed,Codes). if Term is atomic but not an atom

  * TermName
    with compound(Term,TermName,_) if Term is compound (including 0 arity) compounds of SWI-7

?- term_name( atom, Name ).
Name = atom.

?- term_name( 3, Name ).
Name = '3'.

?- term_name( abc(), Name ).
Name = abc.

?- term_name( abc(x), Name ).
Name = abc.

@author nicos angelopoulos
@version  0.1 2016/01/11

*/
term_name( Term, Name ) :-
	atomic( Term ),
	!,
	( atom(Term) -> IsAtom=true; IsAtom=false ),
	term_name_atom( IsAtom, Term, Name ).
term_name( Term, Name ) :-
	compound( Term, Name, _ ).

term_name_atom( true, Atom, Atom ).
term_name_atom( false, Atomic, Atom ) :-
	atom_codes( Atomic, Codes ),
	atom_codes( Atom, Codes ).

%% max_min_list( +List, -Max, -Min ).
%
% Find the maximum and the minimum of a list of numbers in one pass.
%
%@author nicos angelopoulos
%@version  0.1 2014/5/7
%
max_min_list( [H|T], Max, Min ) :-
	max_min_list_1( T, H, H, Max, Min ).

max_min_list_1( [], Max, Min, Max, Min ).
max_min_list_1( [H|T], CurMax, CurMin, Max, Min ) :-
	( H > CurMax ->
		NxMax is H,
		NxMin is CurMin
		;
		NxMax is CurMax,
		( H < CurMin ->
			NxMin is H 
			;
			NxMin is CurMin
		)
	),
	max_min_list_1( T, NxMax, NxMin, Max, Min ).

%% compare_arithmetic( -Op, +X, +Y ).
%
% As compare, but using arithmetic operations. 
%
%==
% ?- compare( Op, 3, 3.0 ).
% Op = (>).
% 
% ?- compare_arithmetic( Op, 3, 3.0 ).
% Op = (=).
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/2/16
%
compare_arithmetic( Op, X, Y ) :-
	( X =:= Y -> 
		Op = =
		;
		( X < Y ->
			Op = <
			;
			Op = >
		)
	).

%% compound( +Term, -Name, -Args ).
%
%  Tries to deal with syntax changes that allow a() as a legal term.
%
% Examples run on Swi.7 
%==
%  compound( abc, Name, Args ).
%  false.
%  
%  compound( abc(a,b,c), Name, Args ).
% Name = abc,
% Args = [a, b, c].
% 
% compound( Term, abc, [a,b,c] ).
% Term = abc(a, b, c).
% 
% compound( Term, abc, [] ).
% Term = abc().
%
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/1/10 (round about)
%
compound( Term, Name, Args ) :-
	current_predicate( compound_name_arguments/3 ),
	!,
	once( (compound(Term) ; (ground(Name),is_list(Args))) ),
	% !,
	compound_name_arguments( Term, Name, Args ).
compound( Term, Name, Args ) :-
	once( (compound(Term) ; (ground(Name),ground(Args))) ),
	Term =.. [Name,Args].

/** tail_list( +List, -Tail ).

Tail is a tail of List. List is a Tail of List.
==
 ?- tail_list( [a,b,c], Tail ).
Tail = [a, b, c] ;
Tail = [b, c] ;
Tail = [c] ;
Tail = [].

==

@author nicos angelopoulos
@version  0.1 2015/12/23
@see kv_list/2 for a use case

*/
tail_list( List, List ).
tail_list( [_H|T], Tail ) :-
	tail_list( T, Tail ).

